#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>

#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "ignorance.h"

gboolean	populate_panel(GtkTreeSelection *sel);
gboolean	verify_form();
gboolean	add_rule_from_form();
gboolean	edit_rule_from_form();
gboolean	del_rule_from_form();
gboolean	ignorance_rule_valid(const gchar *ruletext, gint ruletype);
gboolean	ignorance_rulename_valid(const gchar *rule_name);
gboolean	ignorance_rulescore_valid(const gchar *rule_score);

gboolean
on_levelView_row_activated             (GtkTreeSelection *sel,
										gpointer user_data)
{
	//fprintf(stderr,"Ignorance: Tree row activated\n");
	populate_panel(sel);
  return FALSE;
}


void
on_levelAdd_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
	if(verify_form()){
		if(!add_rule_from_form()){
			fprintf(stderr,"Ignorance: Rule not added\n");
		}//if
	}//if
}


void
on_levelEdit_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
	if(verify_form()){
		if(!edit_rule_from_form()){
			fprintf(stderr,"Ignorance: Rule not modified\n");
		}//if
	}//if
}


void
on_levelDel_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
	if(verify_form()){
		if(!del_rule_from_form()){
			fprintf(stderr,"Ignorance: Rule not deleted\n");
		}//if
	}//if
}

gboolean populate_panel(GtkTreeSelection *sel){
	GtkTreeIter iter;
	GtkTreeModel *model;
	gchar	*levnames,
			*rulenames,
			scorestr[BUFSIZ];
	ignorance_level	*level=NULL;
	ignorance_rule	*rule=NULL;
	GString			*levgs,
					*rulegs;
	
	//fprintf(stderr,"Ignorance: Populating settings panel\n");
	
	    if (gtk_tree_selection_get_selected (sel, &model, &iter)){
			gtk_tree_model_get (model, &iter, LEVEL_COLUMN, &levnames, -1);
			gtk_tree_model_get(model,&iter,RULE_COLUMN,&rulenames,-1);
            
			//fprintf(stderr,"Got rule %s on level %s\n",rulenames,levnames);
			
			if(strlen(rulenames)){
				levgs=g_string_new(levnames);
				rulegs=g_string_new(rulenames);
				level=ignorance_get_level_name(levgs);
				rule=ignorance_level_get_rule(level,rulegs);
				if(!rule){
					fprintf(stderr,"Ignorance: Unable to find rule %s on level %s\n",rulegs->str,level->name->str);
					return FALSE;
				}//if
				//fprintf(stderr,"Got rule %s: %s\n",rule->name->str,(gchar*)(rule->value));
				snprintf(scorestr,BUFSIZ,"%d",rule->score);

				gtk_entry_set_text(GTK_ENTRY(rulename),rulenames);
				gtk_entry_set_text(GTK_ENTRY(score),scorestr);
				gtk_entry_set_text(GTK_ENTRY(filtervalue),(const gchar*)(rule->value));
				
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(im_type_cb),rule->flags & IGNORANCE_APPLY_IM);
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chat_type_cb),rule->flags & IGNORANCE_APPLY_CHAT);
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(username_type_cb),rule->flags & IGNORANCE_APPLY_USER);
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(enterleave_type_cb),rule->flags & IGNORANCE_APPLY_ENTERLEAVE);
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(invite_type_cb),rule->flags & IGNORANCE_APPLY_INVITE);

				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(regex_cb),rule->type & IGNORANCE_RULE_REGEX);
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(enabled_cb),rule->enabled);
				
				g_string_free(levgs,TRUE);
				g_string_free(rulegs,TRUE);
			}//rule
			else{
				gtk_entry_set_text(GTK_ENTRY(rulename),"");
				gtk_entry_set_text(GTK_ENTRY(score),"");
				gtk_entry_set_text(GTK_ENTRY(filtervalue),"");
				
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(im_type_cb),FALSE);
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(chat_type_cb),FALSE);
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(username_type_cb),FALSE);
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(enterleave_type_cb),FALSE);
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(invite_type_cb),FALSE);

				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(regex_cb),FALSE);
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(enabled_cb),FALSE);
			}//level
			
			//fprintf(stderr,"Ignorance: done populating panel\n");
			
			g_free(levnames);
			g_free(rulenames);
        }//if

	return FALSE;
}//populate_panel

gboolean	verify_form(){
	gboolean rv=TRUE;
	gchar	*tmp;
	gint	type=IGNORANCE_RULE_SIMPLETEXT;
	
	//fprintf(stderr,"Ignorance: Verifying settings form content\n");
	
	tmp=(gchar*)gtk_entry_get_text(GTK_ENTRY(filtervalue));
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(regex_cb))){
		type=IGNORANCE_RULE_REGEX;
	}//if
	rv=(rv && ignorance_rule_valid(tmp,type));
	
	if(!rv){
		fprintf(stderr,"Rule invalid: %s\n",tmp);
	}//if
	
	tmp=(gchar*)gtk_entry_get_text(GTK_ENTRY(score));
	rv=(rv && ignorance_rulescore_valid(tmp));
	
	if(!rv){
		fprintf(stderr,"Rule score invalid: %s\n",tmp);
	}//if
	
	tmp=(gchar*)gtk_entry_get_text(GTK_ENTRY(rulename));
	rv=(rv && ignorance_rulename_valid(tmp));
	
	if(!rv){
		fprintf(stderr,"Rule name invalid: %s\n",tmp);
	}//if
		
	return rv;
}//verify_form

gboolean	add_rule_from_form(){
	ignorance_rule *rule=NULL;
	ignorance_level *level=NULL;
	GString	*rule_name,
			*level_name;
	GtkTreeSelection	*sel;
	GtkTreeIter iter,
				childiter;
	GtkTreeModel *model;
	gchar	*rule_value;
	gint	rule_score,
			rule_type=IGNORANCE_RULE_SIMPLETEXT,
			rule_flags=0;
	gboolean	rv=FALSE,
				rule_enable;
	
	sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(levelView));
	
	if (!gtk_tree_selection_get_selected (sel, &model, &iter)){
		return rv;
	}//no level selected

	gtk_tree_model_get (model, &iter, LEVEL_COLUMN, &rule_value, -1);
	
	if(!rule_value){
		return rv;
	}else if(!strlen(rule_value)){
		g_free(rule_value);
		return rv;
	}//no level selected
	
	level_name=g_string_new(rule_value);
	g_free(rule_value);
	
	gtk_tree_model_get(model,&iter,RULE_COLUMN,&rule_value,-1);
	//fprintf(stderr,"Ignorance: getting parent node\n");
	if(strlen(rule_value)){
		//fprintf(stderr,"Ignorance: copying iter to childiter\n");
		childiter=iter;
		//fprintf(stderr,"Ignorance: getting new parent iter\n");
		gtk_tree_model_iter_parent(model,&iter,&childiter);	//get parent node
	}//if a rule is highlighted instead of a base level
	//else{
	//	fprintf(stderr,"Ignorance: got parent");
	//}//parent
	
	//fprintf(stderr,"Ignorance: Freeing rule_value\n");
	g_free(rule_value);
	
	level=ignorance_get_level_name(level_name);

	rule_name=g_string_new(gtk_entry_get_text(GTK_ENTRY(rulename)));
	
	rule=ignorance_level_get_rule(level,rule_name);
	
	if(rule){
		fprintf(stderr,"Ignorance: Rule %s already exists - not adding\n",rule_name->str);
		g_string_free(level_name,TRUE);
		g_string_free(rule_name,TRUE);
		return rv;
	}//rule already exists
	
	rule_score=(gint)strtol(gtk_entry_get_text(GTK_ENTRY(score)),NULL,10);
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(regex_cb))){
		rule_type=IGNORANCE_RULE_REGEX;
	}//regex
	
	rule_value=(gchar*)gtk_entry_get_text(GTK_ENTRY(filtervalue));
	
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(im_type_cb))){
		rule_flags |= IGNORANCE_APPLY_IM;
	}//if
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(chat_type_cb))){
		rule_flags |= IGNORANCE_APPLY_CHAT;
	}//if
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(username_type_cb))){
		rule_flags |= IGNORANCE_APPLY_USER;
	}//if
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(enterleave_type_cb))){
		rule_flags |= IGNORANCE_APPLY_ENTERLEAVE;
	}//if
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(invite_type_cb))){
		rule_flags |= IGNORANCE_APPLY_INVITE;
	}//if
	
	rule_enable=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(enabled_cb));
	
	rule=ignorance_rule_newp(rule_name,rule_type,rule_value,rule_score,rule_flags,rule_enable);
	
	rv=ignorance_level_add_rule(level,rule);
	
	if(rv){
		gtk_tree_store_append(GTK_TREE_STORE(model),&childiter,&iter);
		gtk_tree_store_set(GTK_TREE_STORE(model),&childiter,LEVEL_COLUMN,level_name->str,RULE_COLUMN,rule_name->str,-1);
	}//if
	
	g_string_free(level_name,TRUE);
	g_string_free(rule_name,TRUE);
	
	//fprintf(stderr,"Ignorance: Adding rule from form\n");
	
	return rv;
}//add_rule_from_form

gboolean	edit_rule_from_form(){
	ignorance_level	*level;
	ignorance_rule	*rule;
	GString	*rule_name,
			*level_name;
	GtkTreeSelection	*sel;
	GtkTreeIter iter;
	GtkTreeModel *model;
	gchar	*rule_value;
	gint	rule_score,
			rule_type=IGNORANCE_RULE_SIMPLETEXT,
			rule_flags=0;

	sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(levelView));
	
	if (!gtk_tree_selection_get_selected (sel, &model, &iter)){
		return FALSE;
	}//no level selected

	gtk_tree_model_get (model, &iter, LEVEL_COLUMN, &rule_value, -1);
	
	if(!rule_value){
		return FALSE;
	}else if(!strlen(rule_value)){
		g_free(rule_value);
		return FALSE;
	}//no level selected
	
	level_name=g_string_new(rule_value);
	g_free(rule_value);
	level=ignorance_get_level_name(level_name);

	rule_name=g_string_new(gtk_entry_get_text(GTK_ENTRY(rulename)));
	
	rule=ignorance_level_get_rule(level,rule_name);

	if(!rule){
		fprintf(stderr,"Ignorance: Rule \"%s\" not found on level %s\n",rule_name->str,level_name->str);
		g_string_free(rule_name,TRUE);
		g_string_free(level_name,TRUE);
		return FALSE;
	}//no rule selected / no such rule		
	
	rule_score=(gint)strtol(gtk_entry_get_text(GTK_ENTRY(score)),NULL,10);
	rule->score=rule_score;
	
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(regex_cb))){
		rule_type=IGNORANCE_RULE_REGEX;
	}//regex
	rule->type=rule_type;
	
	rule_value=(gchar*)gtk_entry_get_text(GTK_ENTRY(filtervalue));
	g_free(rule->value);
	rule->value=g_strdup(rule_value);
	
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(im_type_cb))){
		rule_flags |= IGNORANCE_APPLY_IM;
	}//if
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(chat_type_cb))){
		rule_flags |= IGNORANCE_APPLY_CHAT;
	}//if
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(username_type_cb))){
		rule_flags |= IGNORANCE_APPLY_USER;
	}//if
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(enterleave_type_cb))){
		rule_flags |= IGNORANCE_APPLY_ENTERLEAVE;
	}//if
	if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(invite_type_cb))){
		rule_flags |= IGNORANCE_APPLY_INVITE;
	}//if
	
	rule->flags=rule_flags;
	rule->enabled=gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(enabled_cb));
	
	g_string_free(rule_name,TRUE);
	g_string_free(level_name,TRUE);
	
	//fprintf(stderr,"Ignorance: Editing rule from form\n");
	
	return TRUE;
}//edit_rule_from_form

gboolean	del_rule_from_form(){
	ignorance_level	*level;
	ignorance_rule	*rule;
	GString	*rule_name,
			*level_name;
	GtkTreeSelection	*sel;
	GtkTreeIter iter;
	GtkTreeModel *model;
	gchar	*rule_value;
	gboolean	rv=FALSE;
	
	sel=gtk_tree_view_get_selection(GTK_TREE_VIEW(levelView));
	
	if (!gtk_tree_selection_get_selected (sel, &model, &iter)){
		return rv;
	}//no level selected

	gtk_tree_model_get (model, &iter, LEVEL_COLUMN, &rule_value, -1);
	
	if(!rule_value){
		return rv;
	}else if(!strlen(rule_value)){
		g_free(rule_value);
		return rv;
	}//no level selected
	
	level_name=g_string_new(rule_value);
	g_free(rule_value);
	level=ignorance_get_level_name(level_name);

	rule_name=g_string_new(gtk_entry_get_text(GTK_ENTRY(rulename)));
	
	rule=ignorance_level_get_rule(level,rule_name);

	if(!rule){
		fprintf(stderr,"Ignorance: Rule \"%s\" not found on level %s\n",rule_name->str,level_name->str);
		g_string_free(rule_name,TRUE);
		g_string_free(level_name,TRUE);
		return rv;
	}//no rule selected / no such rule		
	
	rv=ignorance_level_remove_rule(level,rule_name);
	
	if(rv){
		//fprintf(stderr,"Ignorance: Preparing to remove %s: %s from tree\n",level_name->str,rule_name->str);
		gtk_tree_store_remove(GTK_TREE_STORE(model),&iter);
	}//if successfully removed
	
	g_string_free(rule_name,TRUE);
	g_string_free(level_name,TRUE);
	
	//fprintf(stderr,"Ignorance: Deleting rule from form\n");
	
	return rv;
}//del_rule_from_form


gboolean	load_form_with_levels(GtkTreeView *tree,GPtrArray	*levels){
	int i,j;
	GtkTreeStore	*store=GTK_TREE_STORE(gtk_tree_view_get_model(tree));
	GtkTreeIter		iter,
					ruleiter;
	ignorance_level	*level;
	ignorance_rule	*rule;
	
	//~ if(store){
		//~ gtk_tree_store_append(store,&iter,NULL);
	//~ }else{
		//~ return FALSE;
	//~ }//if
	
	if(!levels){
		return FALSE;
	}//if
	
	for(i=0;i<levels->len;++i){
		level=(ignorance_level*)g_ptr_array_index(levels,i);
		gtk_tree_store_append(store,&iter,NULL);
		gtk_tree_store_set (store, &iter,
                    LEVEL_COLUMN, level->name->str,
                    RULE_COLUMN, "",
                    -1);
		//~ if(level->rules->len){
			//~ gtk_tree_store_append(store,&ruleiter,&iter);
		//~ }//if
		for(j=0;j<level->rules->len;++j){
			rule=(ignorance_rule*)g_ptr_array_index(level->rules,j);
			gtk_tree_store_append(store,&ruleiter,&iter);
			gtk_tree_store_set(store,&ruleiter,
					LEVEL_COLUMN,level->name->str,
					RULE_COLUMN,rule->name->str,
					-1);
		}//for
		
	}//for
	
	return FALSE;
}//load_form_with_levels


gboolean	ignorance_rule_valid(const gchar *ruletext, gint ruletype){
	gboolean	rv=FALSE;
	regex_t	*reg;
	
	switch(ruletype){
		case IGNORANCE_RULE_SIMPLETEXT:
			rv=(NULL==strchr(ruletext,' '));	//no whitespace allowed yet
			break;
		case IGNORANCE_RULE_REGEX:
			reg=(regex_t*)g_malloc(sizeof(regex_t));
			rv=!((gboolean)regcomp(reg,ruletext,REG_EXTENDED | REG_NOSUB));
			rv=(rv && strlen(ruletext));
			g_free(reg);
			break;
	}//switch
	
	return rv;
}//ignorance_rule_valid

gboolean	ignorance_rulename_valid(const gchar *rule_name){
	gboolean rv=(NULL==strchr(rule_name,' '));	//no whitespace allowed yet
	
	rv=(rv && (strlen(rule_name)));

	return rv;
}//ignorance_rulename_valid


gboolean	ignorance_rulescore_valid(const gchar *rule_score){
	gchar *endptr=NULL;
	gboolean rv=FALSE;
	
	strtol(rule_score,&endptr,10);
	
	if(('\0'!=(*rule_score)) && ('\0'==(*endptr))){
		rv=TRUE;
	}//valid string
	rv=(rv && strlen(rule_score));
	
	return rv;
}//ignorance_rulescore_valid
