#define GAIM_PLUGINS
#include "internal.h"

#include "signals.h"
#include "util.h"

#include "gtkplugin.h"
#include "gtkutils.h"
#include "version.h"
#include "blist.h"
#include "cmds.h"

#include "ignorance.h"
#include "interface.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#define IGNORANCE_PLUGIN_ID "Ignorance"
#define SHORTDESC "Ignorance filter"
#define IGNORANCE_VERSION "1.3.0.1"

//list of tolerance levels, rules, and denizens
static GPtrArray *levels;

//list of commands for unregistering at unload time
static GList	*ignorance_commands;

//
// EXPORTED FUNCTIONS
//
static gboolean plugin_load (GaimPlugin * plugin);
static gboolean plugin_unload(GaimPlugin *plugin);
static void init_plugin (GaimPlugin * plugin);
static GtkWidget* get_config_frame(GaimPlugin *plugin);


//
// PRIVATE FUNCTIONS
//
static gboolean	apply_rule(const GString *username, const GString *text, gint flags);
static gboolean	ignorance_user_match(ignorance_level *il, const GString *username);
static gint		ignorance_get_user_level(const GString *username);
static gint		ignorance_apply_score(gint score, const GString *username,gboolean passthrough);
//static gboolean	ignorance_place_user_index(gint level_index,const GString *username);
static gboolean	ignorance_place_user_name(const GString *level_name,const GString *username);
static ignorance_level*	ignorance_get_level_index(gint index);
static gboolean substitute (GaimAccount * account, const gchar *sender, gchar **message, gint msgflags);			    
static GString* strip_markup(const gchar *text);
static gint	lookup_ampcode(gchar *ampcode);
//static gboolean is_numeric(const gchar *s, gint offset, gint len);
static gboolean load_conf();
static gint buf_get_line(gchar *ibuf, gchar **buf, gint *position, gint len);
static gboolean chat_cb(GaimAccount *account,gchar **sender, gchar **buffer,GaimConversation *chat, void *data);
static gboolean im_cb(GaimAccount *account,gchar **sender, gchar **buffer,gint *flags, void *data);
static gboolean sending_im_cb(GaimAccount *account,GaimConversation *conv,gchar **message, void *data);
static gboolean sending_chat_cb(GaimAccount *account,GaimConversation *conv, gchar **message,void *data);
static gboolean chat_joinleave_cb(GaimConversation *conv, const gchar *name, void *data);
static gboolean chat_invited_cb(GaimAccount *account,const gchar *inviter, const gchar *chat, const gchar *invite_message, const GHashTable *components, void *data);
static gboolean apply_command(GaimAccount *account,gchar **message);
static gboolean ignorance_move_user(GaimAccount *account, gchar **message);
static ignorance_level*	get_level_above(gint index);
static ignorance_level*	get_level_below(gint index);
static gboolean import_zinc_list();
static gboolean import_gaim_list();
static void gaim_buddy_add(gpointer key, gpointer value, gpointer user_data);
static GaimCmdRet	command_cb(GaimConversation *gc,const gchar *cmd,gchar **args,gchar **error,void *data);
static gboolean generate_default_levels();
static GaimCmdRet command_mv(GaimConversation *gc, const gchar **args);
static GaimCmdRet command_ls(GaimConversation *gc, const gchar **args);
static GaimCmdRet command_wl(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_bl(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_rm(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_ri(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_rf(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_rin(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_rfn(GaimConversation *gc,const gchar **args);
static GaimCmdRet command_grep(GaimConversation *gc, const gchar **args);


static gboolean
plugin_load (GaimPlugin * plugin)
{
  void *conv_handle = gaim_conversations_get_handle ();
	GaimCmdId	gcid;
  
  //fprintf(stderr,"Loading plugin %s\n",IGNORANCE_PLUGIN_ID);
	gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Loading plugin %s\n", IGNORANCE_PLUGIN_ID);
	
	load_conf();

  gaim_signal_connect (conv_handle, "receiving-im-msg",
		       plugin, GAIM_CALLBACK (im_cb), NULL);
  gaim_signal_connect (conv_handle, "receiving-chat-msg",
		       plugin, GAIM_CALLBACK (chat_cb), NULL);
  gaim_signal_connect(conv_handle,"writing-im-msg",
				plugin,GAIM_CALLBACK(sending_im_cb),NULL);
	gaim_signal_connect(conv_handle,"writing-chat-msg",
				plugin,GAIM_CALLBACK(sending_chat_cb),NULL);
	gaim_signal_connect(conv_handle,"chat-buddy-joining",
				plugin,GAIM_CALLBACK(chat_joinleave_cb),NULL);
	gaim_signal_connect(conv_handle,"chat-buddy-leaving",
				plugin,GAIM_CALLBACK(chat_joinleave_cb),NULL);
	gaim_signal_connect(conv_handle,"chat-invited",
						plugin,GAIM_CALLBACK(chat_invited_cb),NULL);
	
	//register commands with gaim core
	gcid=gaim_cmd_register("mv","ww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT,NULL,command_cb,"mv user group:  Move a user to a group",NULL);
	ignorance_commands=g_list_append(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("ls","wwwwwwwwww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT | GAIM_CMD_FLAG_ALLOW_WRONG_ARGS,NULL,command_cb,"ls group1 group2 ...:  List all users in each given group",NULL);
	ignorance_commands=g_list_append(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("wl","wwwwwwwwww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT | GAIM_CMD_FLAG_ALLOW_WRONG_ARGS,NULL,command_cb,"wl user1 user2 ...:  Whitelist each given user",NULL);
	ignorance_commands=g_list_append(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("bl","wwwwwwwwww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT | GAIM_CMD_FLAG_ALLOW_WRONG_ARGS,NULL,command_cb,"bl user1 user2 ...:  Blacklist each given user",NULL);
	ignorance_commands=g_list_append(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("rm","wwwwwwwwww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT | GAIM_CMD_FLAG_ALLOW_WRONG_ARGS,NULL,command_cb,"rm user1 user2 ...:  Remove each given user from his current group",NULL);
	ignorance_commands=g_list_append(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("ri","w",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT ,NULL,command_cb,"ri expression:  Add a regex autoignore rule with the given expression to the Default group",NULL);
	ignorance_commands=g_list_append(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("rf","w",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT ,NULL,command_cb,"rf expression:  Add a regex filter rule with the given expression to the Default group",NULL);
	ignorance_commands=g_list_append(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("rin","w",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT ,NULL,command_cb,"rin expression:  Add a regex username autoignore rule with the given expression to the Default group",NULL);
	ignorance_commands=g_list_append(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("rfn","w",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM | GAIM_CMD_FLAG_CHAT ,NULL,command_cb,"rfn expression:  Add a regex username filter rule with the given expression to the Default group",NULL);
	ignorance_commands=g_list_append(ignorance_commands,&gcid);
	gcid=gaim_cmd_register("grep","wwwwwwwwww",GAIM_CMD_P_LOW,GAIM_CMD_FLAG_IM|GAIM_CMD_FLAG_CHAT|GAIM_CMD_FLAG_ALLOW_WRONG_ARGS,NULL,command_cb,"grep userexp group1 group2 ...:  Search each given group's usernames for matches to the regex userexp",NULL);
	ignorance_commands=g_list_append(ignorance_commands,&gcid);
	
  return TRUE;
}

static gboolean
plugin_unload(GaimPlugin *plugin){
	GList	*cursor;
	GaimCmdId	gcid;
	//fprintf(stderr,"Unloading plugin %s\n",IGNORANCE_PLUGIN_ID);
	gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unloading plugin %s\n", IGNORANCE_PLUGIN_ID);
	save_conf();
	
	for(cursor=g_list_first(ignorance_commands);cursor;cursor=g_list_next(cursor)){
		gcid=*((GaimCmdId*)(cursor->data));
		gaim_cmd_unregister(gcid);
	}//for
	
	g_list_free(ignorance_commands);
	
	return TRUE;
}//plugin_unload

static GtkWidget *
get_config_frame(GaimPlugin *plugin)
{
	return create_uiinfo(levels);
}//get_config_frame



static GaimGtkPluginUiInfo ui_info =
{
	get_config_frame
};

static GaimPluginInfo info =
{
	GAIM_PLUGIN_MAGIC,
	GAIM_MAJOR_VERSION,
	GAIM_MINOR_VERSION,
	GAIM_PLUGIN_STANDARD,
	GAIM_GTK_PLUGIN_TYPE,
	0,
	NULL,
	GAIM_PRIORITY_DEFAULT,
	IGNORANCE_PLUGIN_ID,
	N_(SHORTDESC),
  IGNORANCE_VERSION,
  N_("Allows you to manage lists of users with various levels of allowable activity."),
  N_("Allows you to manage lists of users with various levels of allowable activity."),
  "Levi Bard <tak@bard.sytes.net>",
  "http://bard.sytes.net/ignorance",
	plugin_load,
	plugin_unload,
	NULL,
	&ui_info,
	NULL,
	NULL,
	NULL
};

static void
init_plugin (GaimPlugin * plugin)
{
	//fprintf(stderr,"Attempting to load %s plugin\n",IGNORANCE_PLUGIN_ID);
}//init_plugin

GAIM_INIT_PLUGIN (command, init_plugin, info)
 
//
// Private functions
//

//callback for received chat messages
static gboolean chat_cb(GaimAccount *account,gchar **sender, gchar **buffer,
					 	GaimConversation *chat, void *data){
	//fprintf(stderr,"Ignorance: Got chat message\n");
	return substitute(account,*sender,buffer, IGNORANCE_APPLY_CHAT | IGNORANCE_APPLY_USER);
}//chat_cb

//callback for received IMs
static gboolean im_cb(GaimAccount *account,gchar **sender, gchar **buffer,
					 	gint *flags, void *data){
	//fprintf(stderr,"Ignorance: Got IM\n");
	return substitute(account,*sender,buffer, IGNORANCE_APPLY_IM | IGNORANCE_APPLY_USER);
}//im_cb

//this will become obsolete soon
static gboolean sending_im_cb(GaimAccount *account, GaimConversation *conv, gchar **message, void *data){
	GString *msg=strip_markup(*message);
	gboolean rv=apply_command(account,&(msg->str));
		
	if(rv){
		**message='\0';
		gaim_conversation_write(conv, NULL, msg->str, GAIM_MESSAGE_NO_LOG, time(NULL));
	}//+command
	
	g_string_free(msg,TRUE);
	
	return rv;
}//sending_im_cb


//this will become obsolete soon
static gboolean	sending_chat_cb(GaimAccount *account, GaimConversation *conv,gchar **message, void *data){
	GString *msg=strip_markup(*message);
	gboolean rv=apply_command(account,&(msg->str));
		
	if(rv){
		**message='\0';
		gaim_conversation_write(conv, NULL, msg->str, GAIM_MESSAGE_NO_LOG, time(NULL));
	}//+command
	
	g_string_free(msg,TRUE);
	
	return rv;
}//sending_chat_cb

//callback for enter/leave messages
static gboolean chat_joinleave_cb(GaimConversation *conv, const gchar *name, void *data){
	gchar *message=g_strdup(name);
	gboolean rv=substitute(conv->account,name,&message,IGNORANCE_APPLY_ENTERLEAVE | IGNORANCE_APPLY_USER);
	g_free(message);
	
	return rv;
}//chat_joinleave_cb

//callback for chat invites
static gboolean chat_invited_cb(GaimAccount *account,const gchar *inviter, const gchar *chat, const gchar *invite_message, const GHashTable *components, void *data){
	gchar *message=g_strdup(invite_message);
	gboolean rv=substitute(account,inviter,&message,IGNORANCE_APPLY_INVITE | IGNORANCE_APPLY_USER | IGNORANCE_APPLY_CHAT);
	g_free(message);
	
	return rv;
}//chat_invited_cb


//switching callback for all Ignorance commands
static GaimCmdRet	command_cb(GaimConversation *gc,const gchar *cmd,gchar **args,gchar **error,void *data){
	GaimCmdRet	rv=GAIM_CMD_RET_CONTINUE;
	
	//fprintf(stderr,"Ignorance: handling command %s\n",cmd);
	
	switch(*cmd){
		case 'm':
			if('v'==cmd[1]){
				rv=command_mv(gc,(const gchar**)args);
			}// "/mv"
			break;
		case 'l':
			if('s'==cmd[1]){
				rv=command_ls(gc,(const gchar**)args);
			}// "/ls"
			break;
		case 'w':
			if('l'==cmd[1]){
				rv=command_wl(gc,(const gchar**)args);
			}// "/wl"
			break;
		case 'b':
			if('l'==cmd[1]){
				rv=command_bl(gc,(const gchar**)args);
			}// "/bl"
			break;
		case 'r':
			if('m'==cmd[1]){
				rv=command_rm(gc,(const gchar**)args);
			}// "/rm"
			else if('f'==cmd[1]){
				if('n'==cmd[2]){
					rv=command_rfn(gc,(const gchar**)args);
				}else{
					rv=command_rf(gc,(const gchar**)args);
				}// "/rfn"
			}// "/rf"
			else if('i'==cmd[1]){
				if('n'==cmd[2]){
					rv=command_rin(gc,(const gchar**)args);
				}else{
					rv=command_ri(gc,(const gchar**)args);
				}// "/rin"
			}// "/ri"
			break;
		case 'g':
			if(cmd==strstr(cmd,"grep")){
				rv=command_grep(gc,(const gchar**)args);
			}// "/grep"
			break;
	}//switch
	
	return rv;
}//command_cb


//mv username userlevel
static GaimCmdRet command_mv(GaimConversation *gc, const gchar **args){
	GString	*username,
			*levelname;
	gchar	*msgbuf;
	
	if(!args[0]){
		return GAIM_CMD_RET_CONTINUE;
	}else if(!args[1]){
		return GAIM_CMD_RET_CONTINUE;
	}//if
				
	username=g_string_new(gc->account->protocol_id);
	g_string_append(username,args[0]);
	
	levelname=g_string_new(args[1]);
				
	msgbuf=(gchar*)g_malloc(BUFSIZ*sizeof(gchar));
				
	if(ignorance_place_user_name(levelname,username)){
		snprintf(msgbuf,BUFSIZ,"Assigned user %s to %s",username->str,levelname->str);
	}else{
		snprintf(msgbuf,BUFSIZ,"Unable to assign user %s to %s - may already be there",username->str,levelname->str);
	}//if
				
	gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));
			
	g_string_free(username,TRUE);
	g_string_free(levelname,TRUE);
	g_free(msgbuf);

	return GAIM_CMD_RET_OK;
}//command_mv

// Lists the users of given groups
// /ls level1, level2, ...
static GaimCmdRet command_ls(GaimConversation *gc, const gchar **args){
	GString	*levelname,
			*username;
	ignorance_level	*level;
	GList	*cursor;
	gchar	*msgbuf;
	gint	i;
	
	if(!args[0]){
		return GAIM_CMD_RET_CONTINUE;
	}//if
				
	levelname=g_string_new("");
				
	for(i=0;args[i];++i){
		g_string_assign(levelname,args[i]);
		level=ignorance_get_level_name(levelname);
		if(level){
			g_string_append(levelname,":");
			gaim_conversation_write(gc,NULL,levelname->str,GAIM_MESSAGE_NO_LOG,time(NULL));
			for(cursor=g_list_first(level->denizens);cursor;cursor=g_list_next(cursor)){
				username=(GString*)(cursor->data);
				gaim_conversation_write(gc,NULL,username->str,GAIM_MESSAGE_NO_LOG,time(NULL));
			}//for
		}else{
			msgbuf=(gchar*)g_malloc(BUFSIZ*sizeof(gchar));
			snprintf(msgbuf,BUFSIZ,"No such group: %s",levelname->str);
			gaim_conversation_write(gc,NULL,msgbuf,GAIM_MESSAGE_NO_LOG,time(NULL));
			g_free(msgbuf);
		}//if
	}//for
				
	g_string_free(levelname,TRUE);
	return GAIM_CMD_RET_OK;
}//command_ls


// Assigns users to the WL level
// /wl user1, user2, ...
static GaimCmdRet command_wl(GaimConversation *gc,const gchar **args){
	GString	*levelname,
			*username;
	ignorance_level	*level;
	gchar	*msgbuf;
	gint	i;
	
	levelname=g_string_new("WL");
	username=g_string_new("");
	level=ignorance_get_level_name(levelname);
				
	if(level){
		for(i=0;args[i];++i){
			g_string_assign(username,gc->account->protocol_id);
			g_string_append(username,args[i]);
			msgbuf=(gchar*)g_malloc(BUFSIZ*sizeof(gchar));
				
			if(ignorance_place_user_name(levelname,username)){
				snprintf(msgbuf,BUFSIZ,"Assigned user %s to %s",username->str,levelname->str);
			}else{
				snprintf(msgbuf,BUFSIZ,"Unable to assign user %s to %s - may already be there",username->str,levelname->str);
			}//if
				
			gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));
			g_free(msgbuf);
		}//for
	}//if
				
	g_string_free(username,TRUE);
	g_string_free(levelname,TRUE);
				
	return GAIM_CMD_RET_OK;
}//command_wl

// Assigns given users to the BL level
// /bl user1, user2, ...
static GaimCmdRet command_bl(GaimConversation *gc,const gchar **args){
	GString	*levelname,
			*username;
	ignorance_level	*level;
	gchar	*msgbuf;
	gint	i;
	
	levelname=g_string_new("BL");
	username=g_string_new("");
	level=ignorance_get_level_name(levelname);
				
	if(level){
		for(i=0;args[i];++i){
			g_string_assign(username,gc->account->protocol_id);
			g_string_append(username,args[i]);
			msgbuf=(gchar*)g_malloc(BUFSIZ*sizeof(gchar));
				
			if(ignorance_place_user_name(levelname,username)){
				snprintf(msgbuf,BUFSIZ,"Assigned user %s to %s",username->str,levelname->str);
			}else{
				snprintf(msgbuf,BUFSIZ,"Unable to assign user %s to %s - may already be there",username->str,levelname->str);
			}//if
				
			gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));
			g_free(msgbuf);
		}//for
	}//if
				
	g_string_free(username,TRUE);
	g_string_free(levelname,TRUE);
				
	return GAIM_CMD_RET_OK;
}//command_bl


// Removes given users from their currently assigned level
// (effectively moves them to the default level)
// /rm user1, user2, ...
static GaimCmdRet command_rm(GaimConversation *gc, const gchar **args){
	GString	*username;
	ignorance_level	*level;
	gint	i;
	gchar	*msgbuf;
	
	username=g_string_new("");
				
	for(i=0;args[i];++i){
		g_string_assign(username,gc->account->protocol_id);
		g_string_append(username,args[i]);
		level=ignorance_get_level_index(ignorance_get_user_level(username));
		if(level){
			msgbuf=(gchar*)g_malloc(BUFSIZ*sizeof(gchar));
					
			if(ignorance_level_remove_denizen(level,username)){
				snprintf(msgbuf,BUFSIZ,"Successfully removed %s from %s",username->str,level->name->str);
			}else{
				snprintf(msgbuf,BUFSIZ,"Unable to remove %s from %s\n",username->str,level->name->str);
			}//if
			gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));
			g_free(msgbuf);
		}//if
	}//for
				
	g_string_free(username,TRUE);
	return GAIM_CMD_RET_OK;
}//command_rm


static GaimCmdRet	command_ri(GaimConversation *gc, const gchar **args){
	gchar	msgbuf[BUFSIZ];
	ignorance_level	*level;
	ignorance_rule	*rule;
	GString	*rulename;
	
	if(!(*args)){
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	level=ignorance_get_level_index(IGNORANCE_LEVEL_DEFAULT);
	
	if(!level){
		//fprintf(stderr,"Ignorance: Unable to get default level\n");
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to get default level\n");
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	rulename=g_string_new(*args);
	
	rule=ignorance_rule_newp(rulename,IGNORANCE_RULE_REGEX,*args,15,IGNORANCE_APPLY_CHAT | IGNORANCE_APPLY_IM,TRUE);
	if(ignorance_level_add_rule(level,rule)){
		snprintf(msgbuf,BUFSIZ,"Successfully added ignore regex %s",*args);
	}else{
		snprintf(msgbuf,BUFSIZ,"Unable to add ignore regex %s",*args);
	}//if
	
	gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));
	
	g_string_free(rulename,TRUE);
	
	return GAIM_CMD_RET_OK;
}//command_ri

static GaimCmdRet	command_rf(GaimConversation *gc, const gchar **args){
	gchar	msgbuf[BUFSIZ];
	ignorance_level	*level;
	ignorance_rule	*rule;
	GString	*rulename;
	
	if(!(*args)){
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	level=ignorance_get_level_index(IGNORANCE_LEVEL_DEFAULT);
	
	if(!level){
		//fprintf(stderr,"Ignorance: Unable to get default level\n");
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to get default level\n");
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	rulename=g_string_new(*args);
	
	rule=ignorance_rule_newp(rulename,IGNORANCE_RULE_REGEX,*args,1,IGNORANCE_APPLY_CHAT | IGNORANCE_APPLY_IM,TRUE);
	
	if(ignorance_level_add_rule(level,rule)){
		snprintf(msgbuf,BUFSIZ,"Successfully added filter regex %s",*args);
	}else{
		snprintf(msgbuf,BUFSIZ,"Unable to add filter regex %s",*args);
	}//if

	gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));
	
	g_string_free(rulename,TRUE);
	
	return GAIM_CMD_RET_OK;
}//command_rf


static GaimCmdRet	command_rin(GaimConversation *gc, const gchar **args){
	gchar	msgbuf[BUFSIZ];
	ignorance_level	*level;
	ignorance_rule	*rule;
	GString	*rulename;
	
	if(!(*args)){
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	level=ignorance_get_level_index(IGNORANCE_LEVEL_DEFAULT);
	
	if(!level){
		//fprintf(stderr,"Ignorance: Unable to get default level\n");
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to get default level\n");
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	rulename=g_string_new(*args);
	
	rule=ignorance_rule_newp(rulename,IGNORANCE_RULE_REGEX,*args,15,IGNORANCE_APPLY_USER,TRUE);
	if(ignorance_level_add_rule(level,rule)){
		snprintf(msgbuf,BUFSIZ,"Successfully added ignore regex %s",*args);
	}else{
		snprintf(msgbuf,BUFSIZ,"Unable to add ignore regex %s",*args);
	}//if
	
	gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));
	
	g_string_free(rulename,TRUE);
	
	return GAIM_CMD_RET_OK;
}//command_rin

static GaimCmdRet	command_rfn(GaimConversation *gc, const gchar **args){
	gchar	msgbuf[BUFSIZ];
	ignorance_level	*level;
	ignorance_rule	*rule;
	GString	*rulename;
	
	if(!(*args)){
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	level=ignorance_get_level_index(IGNORANCE_LEVEL_DEFAULT);
	
	if(!level){
		//fprintf(stderr,"Ignorance: Unable to get default level\n");
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to get default level\n");
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	rulename=g_string_new(*args);
	
	rule=ignorance_rule_newp(rulename,IGNORANCE_RULE_REGEX,*args,1,IGNORANCE_APPLY_USER,TRUE);
	
	if(ignorance_level_add_rule(level,rule)){
		snprintf(msgbuf,BUFSIZ,"Successfully added filter regex %s",*args);
	}else{
		snprintf(msgbuf,BUFSIZ,"Unable to add filter regex %s",*args);
	}//if

	gaim_conversation_write(gc, NULL, msgbuf, GAIM_MESSAGE_NO_LOG, time(NULL));
	
	g_string_free(rulename,TRUE);
	
	return GAIM_CMD_RET_OK;
}//command_rfn

static GaimCmdRet command_grep(GaimConversation *gc, const gchar **args){
	gchar	msgbuf[BUFSIZ];
	ignorance_level	*level;
	GString	*username,
			*levelname;
	gint	i=0,
			found=0;	
	GList	*users=NULL,
			*cursor;
	
	if(!(args[0])){
		return GAIM_CMD_RET_CONTINUE;
	}//if
	if(!(args[1])){
		return GAIM_CMD_RET_CONTINUE;
	}//if
	
	levelname=g_string_new("");
	
	for(i=1;args[i];++i){
		g_string_assign(levelname,args[i]);
		//fprintf(stderr,"Ignorance: searching level %s\n",levelname->str);
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Searching level %s\n",levelname->str);
		level=ignorance_get_level_name(levelname);
		if(level){
			if(ignorance_level_has_denizen_regex(level,args[0],&users)){
				for(cursor=g_list_first(users);cursor;cursor=g_list_next(cursor)){
					username=(GString*)(cursor->data);
					snprintf(msgbuf,BUFSIZ,"Found user %s in level %s",username->str,level->name->str);
					gaim_conversation_write(gc,NULL,msgbuf,GAIM_MESSAGE_NO_LOG,time(NULL));
					++found;
					(*msgbuf)='\0';
					g_string_free(username,TRUE);
				}//for
				g_list_free(users);
			}//if
		}//if
	}//for
	
	if(!found){
		snprintf(msgbuf,BUFSIZ,"No match for %s",args[0]);
		gaim_conversation_write(gc,NULL,msgbuf,GAIM_MESSAGE_NO_LOG,time(NULL));
	}//if
	
	g_string_free(levelname,TRUE);
	
	return GAIM_CMD_RET_OK;
}//command_grep




// Checks a message for a command, attempts to apply it,
// and prints a confirmation message on the local screen if successful
//
// account is the sender's account
// message is the potential command message
//
// returns true if a command was successfully applied, false otherwise
static gboolean	apply_command(GaimAccount *account,gchar **message){
	gboolean rv=FALSE;
	
	//fprintf(stderr,"Ignorance: checking %s for command\n",message);
	
	if(!(*message)){
		return rv;
	}//if
	
	if('+'!=(**message)){
		return rv;
	}//not a command
	
	switch((*message)[1]){
		case 'm':
			if('v'==(*message)[2]){
				//fprintf(stderr,"Ignorance: +mv\n");
				ignorance_move_user(account,message);
				rv=TRUE;
			}//">mv"
			break;
	}//switch
	
	//fprintf(stderr,"Ignorance: returning %d for command\n",rv);
	
	return rv;
}//apply_command

// parses and executes a move command: +mv someuser somelevel
//
// account is the account of the message sender
// message is the (hopefully valid) move command
//
// returns success/failure
#define IGNORANCE_USER_MOVE_NUMTOKENS 3
static gboolean ignorance_move_user(GaimAccount *account, gchar **message){
	gchar **tokenlist=g_strsplit(*message," ",IGNORANCE_USER_MOVE_NUMTOKENS);
	GString	*username,
			*levelname;
	gboolean	rv;
	gchar	*tmp=(gchar*)g_malloc(BUFSIZ*sizeof(gchar));
	gint	index;
		
	//check right away for valid rule string
	for(index=0;index<IGNORANCE_USER_MOVE_NUMTOKENS;++index){
		if(!tokenlist[index]){
			snprintf(tmp,BUFSIZ,"Invalid mv command: %s",*message);
			//fprintf(stderr,"Ignorance: Invalid command %s\n",message);
			g_free(*message);
			*message=tmp;
			g_strfreev(tokenlist);
			return FALSE;
		}//if
	}//for
	
	username=g_string_new(gaim_account_get_protocol_id(account));
	g_string_append(username,tokenlist[1]);
	levelname=g_string_new(tokenlist[2]);
	
	rv=ignorance_place_user_name(levelname,username);
	
	if(rv){
		snprintf(tmp,BUFSIZ,"Assigned %s to %s",username->str,levelname->str);
	}else{
		snprintf(tmp,BUFSIZ,"Unable to assign %s to %s",username->str,levelname->str);
	}//if
	
	g_free(*message);
	*message=tmp;
	
	g_string_free(username,TRUE);
	g_string_free(levelname,TRUE);
	g_strfreev(tokenlist);
	
	return rv;
}//ignorance_move_user



//
// substitute
// checks an incoming message for rule violation
// account: Account from which the message is being sent
// sender: the name of the sender
// message: Pointer to the message string
// msgflags: message type flags, from ignorance.h
//
// Returns: FALSE if further handling required,
//			TRUE otherwise
static gboolean substitute (GaimAccount *account, const gchar *sender, gchar **message, gint msgflags){
	GString	*username=NULL,
			*text=NULL;
	gboolean	rv=FALSE;

	//fprintf(stderr,"Ignorance: Substituting\n");
 
	if (NULL == message){
		return FALSE;
	}else if (NULL == (*message)){
		return FALSE;
	}//if
	
	//fprintf(stderr,"Ignorance: Got message: %s\n",*message);
	
	username=g_string_new(gaim_account_get_protocol_id(account));
	g_string_append(username,sender);
	text=g_string_new(*message);
	
	//fprintf(stderr,"Ignorance: Preparing to apply rules to message \"%s\"\n",text->str);

	rv=apply_rule(username,text,msgflags);
	
	if(rv){
		//fprintf(stderr,"Ignorance: %s: %s violated!\n", username->str,text->str);
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "%s: %s violated!\n",username->str,text->str);
	//}else{
		//fprintf(stderr,"Ignorance: passed!\n");
	}

	g_string_free(username,TRUE);
	g_string_free(text,TRUE);

	return rv;
}//substitute



//
// strip_markup
// Strips simple HTML markup from a string
// Param text: The marked-up text to be stripped
// Returns: A stripped GString* (that needs to be freed)
//** Free the GStrings! **
//
static GString* strip_markup(const gchar *text){
	GString *tmp=g_string_new(text);
	gint i=0,
		starttag=-1,
		startamp=-1,
		offset=0,
		textlen=tmp->len;
		
	gchar	tmpc[textlen+1];
	
	for(i=0;i<textlen;++i){
		switch(text[i]){
			//<tag>
			case '<':
				if(starttag<0){
					starttag=i;
				}
				break;
			case '>':
				if(starttag>=0){
					g_string_erase(tmp,starttag-offset,(i-starttag)+1);
					offset+=(i-starttag)+1;
					starttag=-1;
				}
				break;

			//&specialcharname;
			case '&':
				startamp=i;
				break;
			case ';':	//replace &specialcharname; with specialchar
				strncpy(tmpc,text+(startamp+1),(i-startamp)-1);
				tmpc[(i-startamp)-1]='\0';
				g_string_erase(tmp,startamp-offset,(i-startamp)+1);
				lookup_ampcode(tmpc);
						
				g_string_insert(tmp,startamp-offset,tmpc);
				offset+=(i-startamp)+1;
                offset-=strlen(tmpc);
				break;
		}//switch
	}//for
	
	return tmp;
}//strip_markup


// 
// lookup_ampcode
// Replaces a &specialcharname; HTML sequence with the character
// that it represents
// Param ampcode: The sequence to be replaced
//				The replacement is stored here
// Returns: Currently always 0
//
static gint	lookup_ampcode(gchar *ampcode){
	if(!strcasecmp(ampcode,"amp")){
		strcpy(ampcode,"&");
	}else if(!strcasecmp(ampcode,"quot")){
		strcpy(ampcode,"\"");
	}else if(!strcasecmp(ampcode,"lt")){
		strcpy(ampcode,"<");
	}else if(!strcasecmp(ampcode,"gt")){
		strcpy(ampcode,">");
	}else if(!strcasecmp(ampcode,"apos")){
		strcpy(ampcode,"'");
	}else{
		strcpy(ampcode,"");
	}//if common ampcodes
	return 0;
}//lookup_ampcode


//~ static gboolean is_numeric(const gchar *s, gint offset, gint len){
	//~ gint i;
	//~ gboolean rv=TRUE;
	
	//~ if(0==len){
		//~ return FALSE;
	//~ }//if
	
	//~ for(i=offset;i<len;++i){
		//~ if(!g_ascii_isdigit(s[i])){
			//~ rv=FALSE;
		//~ }//if
	//~ }//for
	
	//~ return rv;
//~ }//is_numeric

// loads the list of levels, rules, and denizens
// from ~/.gaim/ignorance
//
// also tries to load ~/.zinc/ignore and ~/.zinc/whitelist
//
// returns success/failure
 static gboolean load_conf(){
	 gchar *buf, 
	       *ibuf;
	 gint pnt = 0;
	 gsize size;
	 FILE *conffile=NULL;
	 static ignorance_level *tmplvl=NULL;
	 static ignorance_rule *tmprule=NULL;
	 GString *tmpgs=NULL;
	
	 buf = g_build_filename(gaim_user_dir(), "ignorance", NULL);
	
	 //fprintf(stderr,"Ignorance: Attempting to load conf file %s\n",buf);
	 gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Attempting to load conf file %s\n",buf);
	 
  	levels=g_ptr_array_new();

	 if(!(conffile=fopen(buf,"r"))){
		generate_default_levels();
		import_gaim_list();
		import_zinc_list();
		return FALSE;
	 }//if
	 else{
	 	fclose(conffile);
	 }//else

	 g_file_get_contents(buf, &ibuf, &size, NULL);
	 g_free(buf);
	 if(!ibuf) {
		generate_default_levels();
		import_gaim_list();
		import_zinc_list();

		return FALSE;
	 }//if

	 while(buf_get_line(ibuf, &buf, &pnt, size)) {
		 if((*buf)=='#'){
			 //comment; ignore
		 }else if(strstr(buf,"level")){
			tmplvl=ignorance_level_read(buf);
			if(tmplvl){
				//fprintf(stderr,"Ignorance: Adding level %s\n",tmplvl->name->str);
	 			gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Adding level %s\n",tmplvl->name->str);
				g_ptr_array_add(levels,tmplvl);
			}//valid level
		 }else if(strstr(buf,"rule") && tmplvl){
			tmprule=ignorance_rule_read(buf);
			if(tmprule){
				//fprintf(stderr,"Ignorance: Adding rule %s: %s\n",tmprule->name->str,(gchar*)tmprule->value);
	 			gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Adding rule %s: %s\n",tmprule->name->str,(gchar*)(tmprule->value));
		 		ignorance_level_add_rule(tmplvl,tmprule);
			}//valid rule
		 }else if(tmplvl){//users
			tmpgs=g_string_new(buf);
			//fprintf(stderr,"Ignorance: adding denizen %s\n",buf);
	 		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Adding denizen %s\n",buf);
			if(IGNORANCE_LEVEL_DEFAULT==ignorance_get_user_level(tmpgs)){
				ignorance_level_add_denizen(tmplvl,tmpgs);
			}//if
			g_string_free(tmpgs,TRUE);
		 }//ugly switch
	 }//while
	 
	 import_gaim_list();
	 import_zinc_list();
	
	 return TRUE;
 }//load_conf
 
static gboolean generate_default_levels(){
	ignorance_level *tmplvl;
	ignorance_rule *tmprule;
	GString	*tmpgs;
	 
	 	 	tmplvl=ignorance_level_new();
	 	tmplvl->name=g_string_new("Default");
	 	tmplvl->upper_threshhold=10;
	 	tmplvl->lower_threshhold=-10;
	 	tmplvl->index=100;
	 	tmplvl->allow_passthrough=FALSE;
	 	g_ptr_array_add(levels,tmplvl);
		 
		 tmplvl=ignorance_level_new();
		 tmplvl->name=g_string_new("WL");
		 tmplvl->upper_threshhold=10;
		 tmplvl->lower_threshhold=-10;
		 tmplvl->index=1;
		 tmplvl->allow_passthrough=FALSE;
		 g_ptr_array_add(levels,tmplvl);
		 
 		 tmplvl=ignorance_level_new();
		 tmplvl->name=g_string_new("BL");
		 tmplvl->upper_threshhold=10;
		 tmplvl->lower_threshhold=-10;
		 tmplvl->index=1;
		 tmplvl->allow_passthrough=FALSE;
		 g_ptr_array_add(levels,tmplvl);
		 
		 tmpgs=g_string_new("Everything");
		 tmprule=ignorance_rule_newp(tmpgs,IGNORANCE_RULE_REGEX,".",1,IGNORANCE_APPLY_CHAT | IGNORANCE_APPLY_IM,TRUE);
		 ignorance_level_add_rule(tmplvl,tmprule);
		 g_string_free(tmpgs,TRUE);

		return TRUE;
 }//generate_default_levels


 
gboolean save_conf()
{
	FILE *f;
	gchar	*name,
			tempfilename[BUF_LONG];
	gint fd,
		i;

	name = g_build_filename(gaim_user_dir(), "ignorance", NULL);
	strcpy(tempfilename, name);
	strcat(tempfilename,".XXXXXX");
	fd = g_mkstemp(tempfilename);
	if(fd<0) {
		perror(tempfilename);
		g_free(name);
		return FALSE;
	}//if
	if (!(f = fdopen(fd, "w"))) {
		perror("fdopen");
		close(fd);
		g_free(name);
		return FALSE;
	}//if

	fchmod(fd, S_IRUSR | S_IWUSR);

	for(i=0;i<levels->len;++i){
		ignorance_level_write(g_ptr_array_index(levels,i),f);
	}//for
	
	if(fclose(f)) {
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance",
				   "Error writing to %s: %m\n", tempfilename);
		unlink(tempfilename);
		g_free(name);
		return FALSE;
	}//if
	rename(tempfilename, name);
	g_free(name);
	return TRUE;
}//save_conf


// gets a line from a buffer
//
// ibuf is the buffer to be searched
// buf will be pointed to the beginning of the next line
// position marks the current position in ibuf
// len is the length of ibuf
//
// returns true if there is more to read, false otherwise
static gint buf_get_line(gchar *ibuf, gchar **buf, gint *position, gint len) {
	gint pos = *position, spos = pos;

	if (pos == len)
		return 0;

	while (ibuf[pos++] != '\n') {
		if (pos == len)
			return 0;
	}
	pos--;
	ibuf[pos] = 0;
	*buf = &ibuf[spos];
	pos++;
	*position = pos;
	return 1;
}//buf_get_line

//checks a given level for a given user
static gboolean	ignorance_user_match(ignorance_level *il, const GString *username){
	return (ignorance_level_has_denizen(il,username));
}//ignorance_user_match

//returns the index of the level to which a user belongs, or IGNORANCE_LEVEL_DEFAULT
static gint	ignorance_get_user_level(const GString *username){
	gint i=0;
	ignorance_level	*il=NULL;

	for(i=0;i<levels->len;++i){
		il=(ignorance_level*)g_ptr_array_index(levels,i);
		if(ignorance_user_match(il,username)){
			return il->index;
			//break;
		}//if
	}//for

	if(i<levels->len){
		//fprintf(stderr,"Ignorance: user %s found at level %d\n",username->str,il->index);
		return il->index;
	}//if

	//fprintf(stderr, "Ignorance: user %s not found in %d levels\n",username->str, levels->len);
	return IGNORANCE_LEVEL_DEFAULT;	//default level
}//ignorance_get_user_level

//returns a level specified by index, or NULL
static ignorance_level*	ignorance_get_level_index(gint index){
	gint i=0;
	ignorance_level *il=NULL;

	for(i=0;i<levels->len;++i){
		il=(ignorance_level*)g_ptr_array_index(levels,i);
		if(il->index==index){
			return il;
		}//if
	}//for

	return NULL;
}//ignorance_get_level_index


//returns a level specified by name, or NULL
ignorance_level*	ignorance_get_level_name(const GString *levelname){
	//return g_hash_table_find(levels,hash_find_namedlevel,(const gpointer)levelname);
	gint i=0;
	ignorance_level *il=NULL;

	for(i=0;i<levels->len;++i){
		il=(ignorance_level*)g_ptr_array_index(levels,i);
		if(g_string_equal(levelname,il->name)){
			return il;
		}//if
	}//for

	return NULL;
}//ignorance_get_level_index

// applies a violation score to a user, moving him to a different level if necessary
//
// score is the violation score the user just incurred
// username is the user
// passthrough flags whether the user can be passed through multiple levels
//
// returns the properly decremented score
static gint	ignorance_apply_score(gint score, const GString *username, gboolean passthrough){
	ignorance_level	*userlevel=(ignorance_get_level_index(ignorance_get_user_level(username))),
					*newlevel=NULL;
	gint newscore=0;

	if(!userlevel){
		userlevel=ignorance_get_level_index(IGNORANCE_LEVEL_DEFAULT);
	}//if

	if(!passthrough || (userlevel->allow_passthrough)){
		if(score < userlevel->lower_threshhold){
			newlevel=get_level_below(userlevel->index);
			if(newlevel){
				if(ignorance_level_add_denizen(newlevel,username)){
					if(!((ignorance_level_remove_denizen(userlevel,username)) || userlevel->index==IGNORANCE_LEVEL_DEFAULT)){
						//fprintf(stderr,"Unable to remove user from level %s\n",userlevel->name->str);
	 					gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to remove user from level %s\n",userlevel->name->str);
					}//if
				}//if
			}//if
			newscore=score-userlevel->lower_threshhold;
			//fprintf(stderr,"User placed at level %d\n",(userlevel->index)-1);
	 		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "User placed at level %d\n",(userlevel->index)-1);
		} else if(score > userlevel->upper_threshhold) {
			newlevel=get_level_above(userlevel->index);
			if(newlevel){
				if(ignorance_level_add_denizen(newlevel,username)){
					if(!((ignorance_level_remove_denizen(userlevel,username)) || userlevel->index==IGNORANCE_LEVEL_DEFAULT)){
						//fprintf(stderr,"Unable to remove user from level %s\n",userlevel->name->str);
	 					gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to remove user from level %s\n",userlevel->name->str);
					}//if
				}//if
			}//if
			newscore=score-userlevel->upper_threshhold;
			//fprintf(stderr,"User placed at level %d\n",(userlevel->index)+1);
	 		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "User placed at level %d\n",(userlevel->index)-1);
		}else{
			//fprintf(stderr,"User remains at level %d: \"%s\"\n",userlevel->index,userlevel->name->str);
	 		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "User remains at level %d: \"%s\"\n",userlevel->index,userlevel->name->str);
		}//if
	}else{
		//fprintf(stderr,"Passthrough is %d, userlevel is %d\n",passthrough,userlevel->allow_passthrough);
	}//if

	return newscore;
}//ignorance_apply_score


//~ static gboolean	ignorance_place_user_index(gint level_index, const GString *username){
	//~ gint	current_level=ignorance_get_user_level(username);
	//~ ignorance_level	*newlevel=ignorance_get_level_index(level_index);

	//~ if((NULL!=newlevel) && (current_level!=level_index)){
		//~ if(!(ignorance_level_remove_denizen(ignorance_get_level_index(current_level),username))){
			//~ fprintf(stderr,"Unable to remove user %s from level %d\n",username->str,current_level);
		//~ }//unable to remove
		//~ if(!(ignorance_level_add_denizen(newlevel,username))){	
			//~ fprintf(stderr,"Unable to add user %s to level %d\n",username->str,newlevel->index);
		//~ }//unable to add
		//~ else{
			//~ //fprintf(stderr,"Successfully added user %s to level %d\n",username->str,newlevel->index);
			//~ return TRUE;
		//~ }//success
	//~ }//if
	
	//~ return FALSE;
//~ }//ignorance_place_user_index

//places a given user at a given level, removing him from his current level and returning success/failure
static gboolean	ignorance_place_user_name(const GString *level_name, const GString *username){
	gint	current_level=ignorance_get_user_level(username);
	ignorance_level	*newlevel=ignorance_get_level_name(level_name);

	if((NULL!=newlevel)){
		//fprintf(stderr,"Ignorance: Preparing to move %s from level %d to level %d\n",username->str,current_level,newlevel->index);
		
		if(newlevel->index!=current_level){
			ignorance_level_remove_denizen(ignorance_get_level_index(current_level),username);
			ignorance_level_add_denizen(newlevel,username);	
			return TRUE;
		}//if
	}//if
	else{
		//fprintf(stderr,"Ignorance: Invalid level %s\n",level_name->str);
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Invalid level %s\n",level_name->str);
	}//else
	
	return FALSE;
}//ignorance_place_user_name

// applies a user's message to the rules for his current level
//
// username is the user
// text is the message
// flags are the message type flags
//
// returns true if ok, false on violation
static gboolean	apply_rule(const GString *username, const GString *text, gint flags){
	gint	user_level_index=ignorance_get_user_level(username);
	ignorance_level	*user_level=ignorance_get_level_index(user_level_index);
	gint	text_score=ignorance_level_rulecheck(user_level,text,flags);
	gboolean	passthrough=FALSE;

	//fprintf(stderr,"Ignorance: User %s is currently at level %d\n",username->str,user_level_index);
	
	if(flags & IGNORANCE_APPLY_USER){
		text_score+=ignorance_level_rulecheck(user_level,username,flags);
	}//applies to usernames

	if(!text_score){
		//fprintf(stderr,"%s violated no rules\n",text->str);
		return FALSE;
	} else {
		for(;(text_score=ignorance_apply_score(text_score,username,passthrough));passthrough=TRUE);
		//fprintf(stderr,"%s violated one or more rules\n",text->str);
		return TRUE;
	}//if
}//apply_rule


//gets the level with the greatest index less than the given index, or NULL
static ignorance_level*	get_level_below(gint index){
	gint	i,
		lowerindex=-1;
	ignorance_level	*lev=NULL,
					*cursor=NULL;
	
	for(i=0;i<levels->len;++i){
		cursor=g_ptr_array_index(levels,i);
		if((cursor->index < index) && (cursor->index > lowerindex)){
			lev=cursor;
		}//if
	}//for
	
	return lev;
}//get_level_below


//gets the level with the least index greater than the given index, or NULL
static ignorance_level*	get_level_above(gint index){
	gint	i,
		upperindex=INT_MAX;
	ignorance_level	*lev=NULL,
					*cursor=NULL;
	
	for(i=0;i<levels->len;++i){
		cursor=g_ptr_array_index(levels,i);
		if((cursor->index > index) && (cursor->index < upperindex)){
			lev=cursor;
		}//if
	}//for
	
	return lev;
}//get_level_above

//attemps to import ignore list and whitelist from ~/.zinc
static gboolean import_zinc_list(){
	gchar	*buf,
			*ibuf;
	gint	pnt=0;
	gsize	size=0;
	FILE	*zincfile;
	ignorance_level	*zinclevel=NULL;
	ignorance_rule	*zincrule=NULL;
	GString	*tmp=NULL;
	
	buf=g_build_filename(g_get_home_dir(),".zinc","ignore", NULL);
	
	//fprintf(stderr,"Ignorance: Attempting to open %s\n",buf);
	
	if(!(zincfile=fopen(buf,"r"))){
		//fprintf(stderr,"Ignorance: Unable to open %s\n",buf);
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to open %s\n",buf);
		g_free(buf);
		return FALSE;
	}else{
		fclose(zincfile);
	}//if
	
	//fprintf(stderr,"Ignorance: Reading %s\n",buf);
	g_file_get_contents(buf, &ibuf, &size, NULL);
	
	tmp=g_string_new("ZincBL");
	zinclevel=ignorance_get_level_name(tmp);
	if(!zinclevel){
		//fprintf(stderr,"Ignorance: Creating new Zinc blacklist\n");
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Creating new Zinc blacklist\n");
		zincrule=ignorance_rule_newp(g_string_new("Everything"),IGNORANCE_RULE_REGEX,(gchar*)".",1,(IGNORANCE_APPLY_CHAT | IGNORANCE_APPLY_IM | IGNORANCE_APPLY_ENTERLEAVE),TRUE);
				
		zinclevel=ignorance_level_new();
		zinclevel->name=g_string_new(tmp->str);
		zinclevel->index=999;
		ignorance_level_add_rule(zinclevel,zincrule);
		g_ptr_array_add(levels,zinclevel);
	}//if no ZincBL already
	//else{
		//fprintf(stderr,"Ignorance: Using existing Zinc blacklist\n");
	//}//else
	
	if(!tmp){
		tmp=g_string_new("");
	}//if
	
	//fprintf(stderr,"Ignorance: Preparing to read in zinc blacklist users\n");
	gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Preparing to read in zinc blacklist users\n");
	//fprintf(stderr,"%s %d\n",ibuf,size);
	for(;buf_get_line(ibuf, &buf, &pnt, size);){
		//fprintf(stderr,"Ignorance: Got user %s\n",buf);
		g_string_assign(tmp,"prpl-yahoo");
		g_string_append(tmp,buf);
		if(IGNORANCE_LEVEL_DEFAULT==ignorance_get_user_level(tmp)){
			ignorance_level_add_denizen(zinclevel,tmp);
		}//if
	}//for
	
	pnt=0;
	buf=g_build_filename(g_get_home_dir(),".zinc","whitelist", NULL);
	
	//fprintf(stderr,"Ignorance: Reading %s\n",buf);
	g_file_get_contents(buf, &ibuf, &size, NULL);
	
	g_string_assign(tmp,"ZincWL");
	zinclevel=ignorance_get_level_name(tmp);
	if(!zinclevel){
		//fprintf(stderr,"Ignorance: Creating new Zinc whitelist\n");
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Creating new Zinc whitelist\n");
		zinclevel=ignorance_level_new();
		zinclevel->name=g_string_new(tmp->str);
		zinclevel->index=2;
		g_ptr_array_add(levels,zinclevel);
	}//if no ZincWL already
	//else{
		//fprintf(stderr,"Ignorance: Using existing Zinc whitelist\n");
	//}//else
	
	//fprintf(stderr,"Ignorance: Preparing to read in zinc whitelist users\n");
	gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Preparing to read in zinc whitelist users\n");
	for(;buf_get_line(ibuf, &buf, &pnt, size);){
		//fprintf(stderr,"Ignorance: Got user %s\n",buf);
		g_string_assign(tmp,"prpl-yahoo");
		g_string_append(tmp,buf);
		if(IGNORANCE_LEVEL_DEFAULT==ignorance_get_user_level(tmp)){
			ignorance_level_add_denizen(zinclevel,tmp);
		}//if
	}//for
	
	//fprintf(stderr,"Ignorance: Done importing Zinc users\n");
	gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Done importing Zinc users\n");
	
	return TRUE;
}//import_zinc_list


static gboolean import_gaim_list(){
	GaimBuddyList *bl=gaim_get_blist();	
	GString	*wlname=g_string_new("WL");
	ignorance_level	*wl=ignorance_get_level_name(wlname);
	gboolean	rv=FALSE;
	
	//fprintf(stderr,"Ignorance: Loading Gaim buddy list\n");
	
	if(bl && wl){
		g_hash_table_foreach(bl->buddies,gaim_buddy_add,wl);
		rv=TRUE;
	}else{
		//fprintf(stderr,"Ignorance: Unable to get Gaim buddy list!\n");
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Unable to get Gaim buddy list!\n");
	}//if
	
	g_string_free(wlname,TRUE);
	
	return rv;
}//import_gaim_list

static void gaim_buddy_add(gpointer key, gpointer value, gpointer user_data){
	GaimBuddy	*buddy=(GaimBuddy*)value;
	ignorance_level	*level=(ignorance_level*)user_data;
	GString	*tmp;
	
	if(buddy && level){
		tmp=g_string_new(buddy->account->protocol_id);
		g_string_append(tmp,buddy->name);
		
		//fprintf(stderr,"Ignorance: Got Gaim buddy %s\n",tmp->str);
		
		if(IGNORANCE_LEVEL_DEFAULT==ignorance_get_user_level(tmp)){
			ignorance_level_add_denizen(level,tmp);
		//}else{
			//fprintf(stderr,"Ignorance: skipping %s: already exists\n",tmp->str);	
		}//if
		
		
		
		g_string_free(tmp,TRUE);
	}else{
		//fprintf(stderr,"Ignorance: Bad arguments to gaim_buddy_add\n");
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Bad arguments to gaim_buddy_add\n");
	}//if
}//gaim_buddy_add
