#include "ignorance_rule.h"
#include <string.h>
#include <strings.h>



gboolean	assign_rule_token(ignorance_rule *rule,const char *tokentxt);

 ignorance_rule*	ignorance_rule_new(){
	ignorance_rule *ir=(ignorance_rule*)g_malloc(sizeof(ignorance_rule));

	ir->name=g_string_new("");
	ir->type=IGNORANCE_RULE_SIMPLETEXT;
	ir->value=(gpointer)g_string_new("");
	ir->score=0;
	ir->flags=0;
	ir->enabled=TRUE;

	return ir;
}//ignorance_rule_new



 ignorance_rule*	ignorance_rule_newp(const GString *name,
					gint type,
					const gchar *value,
					gint score, 
					gint flags,
					gboolean enabled){
	ignorance_rule *ir=(ignorance_rule*)g_malloc(sizeof(ignorance_rule));

	ir->name=g_string_new(name->str);
	if(ignorance_rule_has_type(type)){
		ir->type=type;
	} else {
		ir->type=IGNORANCE_RULE_INVALID;
	}//if
	ir->value=g_strdup(value);
	ir->score=score;
	ir->flags=flags;
	ir->enabled=enabled;

	return ir;
}//ignorance_rule_newp


 void	ignorance_rule_free(ignorance_rule *ir){
	g_string_free(ir->name,TRUE);
	g_free(ir->value);
	g_free(ir);
}//ignorance_rule_free


 void	ignorance_rule_free_g(gpointer ir,gpointer user_data){
	ignorance_rule_free((ignorance_rule*)ir);
}//ignorance_rule_free_g


 gboolean	ignorance_rule_has_type(gint type){
	if((type>=IGNORANCE_RULE_MINVALID) || (type<=IGNORANCE_RULE_INVALID))
		return FALSE;
	return TRUE;
}//ignorance_rule_has_type


 gint	ignorance_rule_rulecheck(ignorance_rule *rule,const GString *text, gint flags){
	if((flags & rule->flags) && rule->enabled){//match flag types
		switch(rule->type){
			case IGNORANCE_RULE_SIMPLETEXT:
				//fprintf(stderr,"Ignorance: Preparing to check %s against %s\n",text->str,rule->name->str);
				return simple_text_rulecheck(rule,text);
			case IGNORANCE_RULE_REGEX:
				//fprintf(stderr,"Ignorance: Preparing to check %s against regex %s\n",text->str,rule->name->str);
				return regex_rulecheck(rule,text);
			default:
				return 0;
		}//switch
	}//if

	return 0;
}//ignorance_rule_rulecheck



//obsolete?
//ignorance_rule*	ignorance_rule_parse(const gchar *ruletext){
//	return simple_text_ruleparse(ruletext);
//}//ignorance_rule_parse


 gint	simple_text_rulecheck(ignorance_rule *rule,const GString *text){
	const gchar *rulevalue=(const gchar*)(rule->value);
	 
	if(NULL!=g_strstr_len(text->str,text->len,rulevalue)){
		return rule->score;
	}//if

	return 0;
}//simple_text_rulecheck


 //~ ignorance_rule*	simple_text_ruleparse(const gchar *ruletext){
	//~ ignorance_rule *ir=ignorance_rule_new();
	//~ gchar **tokenlist=g_strsplit(ruletext,";",IGNORANCE_RULE_SIMPLETEXT_NUMTOKENS);
	//~ gint index=0;
	
	//~ //check right away for valid rule string
	//~ for(index=0;index<IGNORANCE_RULE_SIMPLETEXT_NUMTOKENS;++index){
		//~ if(!tokenlist[index]){
			//~ fprintf(stderr,"Ignorance: Invalid rule %s\n",ruletext);
			//~ g_strfreev(tokenlist);
			//~ ignorance_rule_free(ir);
			//~ return NULL;
		//~ }//if
	//~ }//for
	
	//~ ir->name=g_string_new((gchar*)tokenlist[0]);
	//~ ir->type=strtol((gchar*)tokenlist[1],NULL,10);

	//~ //assuming string for now
	//~ ir->value=g_strdup((gchar*)tokenlist[2]);
	//~ ir->score=strtol((gchar*)tokenlist[3],NULL,10);
	//~ ir->flags=strtol((gchar*)tokenlist[4],NULL,10);
	
	//~ //if(!ignorance_rule_has_type(ir->type)){
	//~ if(ir->type!=IGNORANCE_RULE_SIMPLETEXT){
		//~ //fprintf(stderr,"Ignorance: Invalid type %d\n",ir->type);
		//~ g_strfreev(tokenlist);
		//~ ignorance_rule_free(ir);
		//~ return NULL;
	//~ }//if
	
	//~ //fprintf(stderr,"Ignorance: Parsed simpletext rule\n");
	
	//~ g_strfreev(tokenlist);
	//~ return ir;
//~ }//simple_text_ruleparse


 gint	regex_rulecheck(ignorance_rule *rule, const GString *text){
	regex_t	*reg=(regex_t*)g_malloc(sizeof(regex_t));
	gint rv=0;
	 
	 //fprintf(stderr,"Ignorance: Preparing to check %s against '%s'\n",text->str,(const gchar*)rule->value);
	
	if(regcomp(reg,(const gchar*)rule->value,REG_EXTENDED | REG_NOSUB)){
		//fprintf(stderr,"Ignorance: Error parsing regex %s\n",(const gchar*)rule->value);
		gaim_debug(GAIM_DEBUG_ERROR, "ignorance", "Error parsing regex %s\n",(const gchar*)(rule->value));
	}else if(regexec(reg,text->str,1,NULL,0)){
		//fprintf(stderr,"Ignorance: No match for regex %s on text %s\n",(const gchar*)rule->value,text->str);
	}else{
		//fprintf(stderr,"Ignorance: Regex %s matched text %s\n",(const gchar*)rule->value,text->str);
		rv=rule->score;
	}//if
		
	g_free(reg);
	return rv;
}//regex_rulecheck


 //~ ignorance_rule*	regex_ruleparse(const gchar *ruletext){
	//~ ignorance_rule *ir=ignorance_rule_new();
	//~ gchar **tokenlist=g_strsplit(ruletext,";",IGNORANCE_RULE_REGEX_NUMTOKENS);
	//~ gint index=0;
	
	//~ //check right away for valid rule string
	//~ for(index=0;index<IGNORANCE_RULE_REGEX_NUMTOKENS;++index){
		//~ if(!tokenlist[index]){
			//~ fprintf(stderr,"Ignorance: Invalid rule %s\n",ruletext);
			//~ g_strfreev(tokenlist);
			//~ ignorance_rule_free(ir);
			//~ return NULL;
		//~ }//if
	//~ }//for
	
	//~ ir->name=g_string_new((gchar*)tokenlist[0]);
	//~ ir->type=strtol((gchar*)tokenlist[1],NULL,10);

	//~ //assuming string for now
	//~ ir->value=g_strdup((gchar*)tokenlist[2]);
	//~ ir->score=strtol((gchar*)tokenlist[3],NULL,10);
	//~ ir->flags=strtol((gchar*)tokenlist[4],NULL,10);
	
	//~ //if(!ignorance_rule_has_type(ir->type)){
	//~ if(ir->type!=IGNORANCE_RULE_REGEX){
		//~ fprintf(stderr,"Ignorance: Invalid type %d\n",ir->type);
		//~ g_strfreev(tokenlist);
		//~ ignorance_rule_free(ir);
		//~ return NULL;
	//~ }//if
	
	//~ //fprintf(stderr,"Ignorance: parsed regex rule\n");
	
	//~ g_strfreev(tokenlist);
	//~ return ir;
//~ }//regex_ruleparse


ignorance_rule*	ignorance_rule_read(const gchar *ruletext){
	gchar	*tokptr=strtok((gchar*)ruletext," ");
	ignorance_rule *rule=ignorance_rule_new();

	if(!tokptr){
		ignorance_rule_free(rule);
		return NULL;
	}//if


	for(tokptr=strtok(NULL," ");tokptr;tokptr=strtok(NULL," ")){
		assign_rule_token(rule,tokptr);
	}//for

	return rule;
}//ignorance_level_read



// Parses out a token of the form tokenname="value" and assigns it to a rulename
// 
// rule is the rule to be updated
// tokentxt is the token string
// true returned if token is valid and successfully added to the rule
//
//level name1="value1" name2="value2" ...
gboolean	assign_rule_token(ignorance_rule *rule,const gchar *tokentxt){
	gchar	*name=(gchar*)tokentxt,
		*value=NULL;
	gboolean rv=TRUE;
	gint	cursor=0;

		value=strchr(tokentxt,'=');
		if(value){	
			(*value)='\0';
			++value;

			//strip "value" quotes
			if('"'==(*value)){
				++value;
				cursor=strlen(value)-1;
				if('"'==value[cursor]){
					value[cursor]='\0';
				}//if
			}//if

			//HACK!
			if(!strncasecmp(name,"name",BUFSIZ)){
				g_string_assign(rule->name,value);
			}else if(!strncasecmp(name,"type",BUFSIZ)){
				rule->type=atoi(value);
			}else if(!strncasecmp(name,"value",BUFSIZ)){
				rule->value=(gchar*)g_malloc((strlen(value)+1)*sizeof(gchar));
				strncpy(rule->value,value,strlen(value)+1);
			}else if(!strncasecmp(name,"score",BUFSIZ)){
				rule->score=atoi(value);
			}else if(!strncasecmp(name,"flags",BUFSIZ)){
				rule->flags=atoi(value);
			}else if(!strncasecmp(name,"enabled",BUFSIZ)){
				rule->enabled=(gboolean)atoi(value);
			}else{
				rv=FALSE;
			}//long, ugly switch
		}else{
			rv=FALSE;
		}//value

		return rv;
}//assign_rule_token


gboolean	ignorance_rule_write(ignorance_rule *rule,FILE *f){
	fprintf(f,"rule name=\"%s\" type=\"%d\" score=\"%d\" value=\"%s\" flags=\"%d\" enabled=\"%d\"\n",
			rule->name->str,rule->type,rule->score,(gchar*)(rule->value),rule->flags,rule->enabled);
			
	return TRUE;
}//ignorance_rule_write
